from typing import AbstractSet

from ..data.models.room_rank import all_four_player_room_rank, \
    all_three_player_room_rank, \
    all_four_player_south_room_rank, all_four_player_east_room_rank, all_three_player_south_room_rank, \
    all_three_player_east_room_rank, RoomRank, all_three_player_golden_room_rank, all_four_player_golden_room_rank, \
    all_four_player_jade_room_rank, all_three_player_jade_room_rank, all_four_player_throne_room_rank, \
    all_three_player_throne_room_rank


def map_room_rank(room_rank: AbstractSet[RoomRank]) -> str:
    ans = []

    if room_rank & all_four_player_room_rank == all_four_player_room_rank:
        ans.append("四麻整体")
        room_rank = room_rank - all_four_player_room_rank

    if room_rank & all_three_player_room_rank == all_three_player_room_rank:
        ans.append("三麻整体")
        room_rank = room_rank - all_three_player_room_rank

    if room_rank & all_four_player_south_room_rank == all_four_player_south_room_rank:
        ans.append("四人南整体")
        room_rank = room_rank - all_four_player_south_room_rank

    if room_rank & all_four_player_east_room_rank == all_four_player_east_room_rank:
        ans.append("四人东整体")
        room_rank = room_rank - all_four_player_east_room_rank

    if room_rank & all_three_player_south_room_rank == all_three_player_south_room_rank:
        ans.append("三人南整体")
        room_rank = room_rank - all_three_player_south_room_rank

    if room_rank & all_three_player_east_room_rank == all_three_player_east_room_rank:
        ans.append("三人东整体")
        room_rank = room_rank - all_three_player_east_room_rank

    if room_rank & all_four_player_golden_room_rank == all_four_player_golden_room_rank:
        ans.append("四麻金之间整体")
        room_rank = room_rank - all_four_player_golden_room_rank

    if room_rank & all_three_player_golden_room_rank == all_three_player_golden_room_rank:
        ans.append("三麻金之间整体")
        room_rank = room_rank - all_three_player_golden_room_rank

    if room_rank & all_four_player_jade_room_rank == all_four_player_jade_room_rank:
        ans.append("四麻玉之间整体")
        room_rank = room_rank - all_four_player_jade_room_rank

    if room_rank & all_three_player_jade_room_rank == all_three_player_jade_room_rank:
        ans.append("三麻玉之间整体")
        room_rank = room_rank - all_three_player_jade_room_rank

    if room_rank & all_four_player_throne_room_rank == all_four_player_throne_room_rank:
        ans.append("四麻王座之间整体")
        room_rank = room_rank - all_four_player_throne_room_rank

    if room_rank & all_three_player_throne_room_rank == all_three_player_throne_room_rank:
        ans.append("三麻王座之间整体")
        room_rank = room_rank - all_three_player_throne_room_rank

    for rl in room_rank:
        if rl == RoomRank.four_player_golden_south:
            ans.append("四麻金南")
        elif rl == RoomRank.four_player_golden_east:
            ans.append("四麻金东")
        elif rl == RoomRank.four_player_jade_south:
            ans.append("四麻玉南")
        elif rl == RoomRank.four_player_jade_east:
            ans.append("四麻玉东")
        elif rl == RoomRank.four_player_throne_south:
            ans.append("四麻王座南")
        elif rl == RoomRank.four_player_throne_east:
            ans.append("四麻王座东")
        if rl == RoomRank.three_player_golden_south:
            ans.append("三麻金南")
        elif rl == RoomRank.three_player_golden_east:
            ans.append("三麻金东")
        elif rl == RoomRank.three_player_jade_south:
            ans.append("三麻玉南")
        elif rl == RoomRank.three_player_jade_east:
            ans.append("三麻玉东")
        elif rl == RoomRank.three_player_throne_south:
            ans.append("三麻王座南")
        elif rl == RoomRank.three_player_throne_east:
            ans.append("三麻王座东")

    return "&".join(ans)



from ..data.models.player_num import PlayerNum

_PLAYER_NUM_MAPPING = {
    PlayerNum.four: "四麻",
    PlayerNum.three: "三麻"
}


def map_player_num(player_num: PlayerNum):
    return _PLAYER_NUM_MAPPING[player_num]



"""
nonebot-plugin-majsoul

@Author         : ssttkkl
@License        : AGPLv3
@GitHub         : https://github.com/ssttkkl/nonebot-plugin-majsoul
"""
from nonebot import require

require("nonebot_plugin_saa")

from .config import Config
from .utils.nonebot import default_cmd_start

__usage__ = f"""
牌谱屋：
- {default_cmd_start}雀魂(三麻)信息 <雀魂账号> [<房间类型>] [最近<数量>场] [最近<数量>{{天|周|个月|年}}]
- {default_cmd_start}雀魂(三麻)对局 <雀魂账号> [<房间类型>]
- {default_cmd_start}雀魂(三麻)PT图 <雀魂账号> [最近<数量>场] [最近<数量>{{天|周|个月|年}}]

牌谱下载：
- {default_cmd_start}下载雀魂牌谱 <牌谱链接或UUID>

以上命令格式中，以<>包裹的表示一个参数，以[]包裹的表示一个可选项。

详细说明：参见https://github.com/bot-ssttkkl/nonebot-plugin-majsoul
""".strip()

from nonebot.plugin import PluginMetadata

__plugin_meta__ = PluginMetadata(
    name='雀魂查询',
    description='根据牌谱屋的数据查询雀魂账号信息',
    usage=__usage__,
    type="application",
    config=Config,
    homepage="https://github.com/bot-ssttkkl/nonebot-plugin-majsoul",
    supported_adapters={"~onebot.v11", "~onebot.v12", "~qqguild", "~telegram", "~kaiheila"}
)

from . import paifuya
from . import paipu



import asyncio
from functools import wraps
from typing import Type

from httpx import HTTPError
from nonebot import logger
from nonebot.exception import MatcherException, ActionFailed
from nonebot.internal.matcher import Matcher

from nonebot_plugin_majsoul.errors import BadRequestError


def handle_error(matcher: Type[Matcher], silently: bool = False):
    def decorator(func):
        @wraps(func)
        async def wrapped_func(*args, **kwargs):
            try:
                return await func(*args, **kwargs)
            except MatcherException as e:
                raise e
            except BadRequestError as e:
                if not silently:
                    await matcher.finish(e.message)
            except ActionFailed as e:
                # 避免当发送消息错误时再尝试发送
                logger.exception(e)
            except HTTPError as e:
                logger.exception(e)
                if not silently:
                    await matcher.finish(f"网络错误")
            except asyncio.TimeoutError as e:
                logger.warning(type(e))
                if not silently:
                    await matcher.finish(f"查询超时")
            except Exception as e:
                logger.exception(e)
                if not silently:
                    await matcher.finish(f"内部错误：{type(e)}{str(e)}")

        return wrapped_func

    return decorator



from typing import List, Optional

from pydantic import BaseModel, Extra

from .player_info import PlayerLevel
from .room_rank import RoomRank


class PlayerStats(BaseModel):
    id: int
    nickname: str
    count: int
    played_modes: List[RoomRank]
    level: PlayerLevel
    max_level: PlayerLevel
    rank_rates: List[float]
    rank_avg_score: List[Optional[float]]
    avg_rank: float
    # negative_rate: float

    class Config:
        extra = Extra.ignore



import asyncio
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from os import cpu_count

my_executor = ThreadPoolExecutor(cpu_count())


async def run_in_my_executor(func, *args, **kwargs):
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(my_executor, partial(func, *args, **kwargs))



from typing import AbstractSet, Tuple, Optional

from ..data.models.room_rank import RoomRank, all_four_player_south_room_rank, \
    all_four_player_east_room_rank, all_three_player_east_room_rank, all_three_player_south_room_rank, \
    all_four_player_golden_room_rank, all_three_player_golden_room_rank, all_four_player_jade_room_rank, \
    all_three_player_jade_room_rank, all_four_player_throne_room_rank, all_three_player_throne_room_rank

_ROOM_RANK_REVERSED_MAPPING = {
    "南": (all_four_player_south_room_rank, all_three_player_south_room_rank),
    "东": (all_four_player_east_room_rank, all_three_player_east_room_rank),

    "金": (all_four_player_golden_room_rank, all_three_player_golden_room_rank),
    "玉": (all_four_player_jade_room_rank, all_three_player_jade_room_rank),
    "王座": (all_four_player_throne_room_rank, all_three_player_throne_room_rank),

    "金东": ({RoomRank.four_player_golden_east}, {RoomRank.three_player_golden_east}),
    "金南": ({RoomRank.four_player_golden_south}, {RoomRank.three_player_golden_south}),
    "玉东": ({RoomRank.four_player_jade_east}, {RoomRank.three_player_jade_east}),
    "玉南": ({RoomRank.four_player_jade_south}, {RoomRank.three_player_jade_south}),
    "王座东": ({RoomRank.four_player_throne_east}, {RoomRank.three_player_throne_east}),
    "王座南": ({RoomRank.four_player_throne_south}, {RoomRank.three_player_throne_south}),
}

_ROOM_RANK_REVERSED_MAPPING["南场"] = _ROOM_RANK_REVERSED_MAPPING["南"]
_ROOM_RANK_REVERSED_MAPPING["东场"] = _ROOM_RANK_REVERSED_MAPPING["东"]
_ROOM_RANK_REVERSED_MAPPING["金之间"] = _ROOM_RANK_REVERSED_MAPPING["金"]
_ROOM_RANK_REVERSED_MAPPING["玉之间"] = _ROOM_RANK_REVERSED_MAPPING["玉"]
_ROOM_RANK_REVERSED_MAPPING["王座之间"] = _ROOM_RANK_REVERSED_MAPPING["王座"]
_ROOM_RANK_REVERSED_MAPPING["金之间东场"] = _ROOM_RANK_REVERSED_MAPPING["金东"]
_ROOM_RANK_REVERSED_MAPPING["玉之间东场"] = _ROOM_RANK_REVERSED_MAPPING["玉东"]
_ROOM_RANK_REVERSED_MAPPING["王座之间东场"] = _ROOM_RANK_REVERSED_MAPPING["王座东"]
_ROOM_RANK_REVERSED_MAPPING["金之间南场"] = _ROOM_RANK_REVERSED_MAPPING["金南"]
_ROOM_RANK_REVERSED_MAPPING["玉之间南场"] = _ROOM_RANK_REVERSED_MAPPING["玉南"]
_ROOM_RANK_REVERSED_MAPPING["王座之间南场"] = _ROOM_RANK_REVERSED_MAPPING["王座南"]


def try_parse_room_rank(raw: str) -> Optional[Tuple[AbstractSet[RoomRank], AbstractSet[RoomRank]]]:
    return _ROOM_RANK_REVERSED_MAPPING.get(raw, None)



from enum import Enum


class RoomRank(int, Enum):
    four_player_golden_east = 8
    four_player_golden_south = 9
    four_player_jade_east = 11
    four_player_jade_south = 12
    four_player_throne_east = 15
    four_player_throne_south = 16
    three_player_golden_east = 21
    three_player_golden_south = 22
    three_player_jade_east = 23
    three_player_jade_south = 24
    three_player_throne_east = 25
    three_player_throne_south = 26


all_four_player_east_room_rank = frozenset([
    RoomRank.four_player_golden_east,
    RoomRank.four_player_jade_east,
    RoomRank.four_player_throne_east,
])

all_four_player_south_room_rank = frozenset([
    RoomRank.four_player_golden_south,
    RoomRank.four_player_jade_south,
    RoomRank.four_player_throne_south,
])

all_four_player_room_rank = frozenset([
    *all_four_player_east_room_rank,
    *all_four_player_south_room_rank
])

all_three_player_east_room_rank = frozenset([
    RoomRank.three_player_golden_east,
    RoomRank.three_player_jade_east,
    RoomRank.three_player_throne_east,
])

all_three_player_south_room_rank = frozenset([
    RoomRank.three_player_golden_south,
    RoomRank.three_player_jade_south,
    RoomRank.three_player_throne_south,
])

all_three_player_room_rank = frozenset([
    *all_three_player_east_room_rank,
    *all_three_player_south_room_rank
])

all_four_player_golden_room_rank = frozenset([
    RoomRank.four_player_golden_east, RoomRank.four_player_golden_south
])
all_three_player_golden_room_rank = frozenset([
    RoomRank.three_player_golden_east, RoomRank.three_player_golden_south
])

all_four_player_jade_room_rank = frozenset([
    RoomRank.four_player_jade_east, RoomRank.four_player_jade_south
])
all_three_player_jade_room_rank = frozenset([
    RoomRank.three_player_jade_east, RoomRank.three_player_jade_south
])

all_four_player_throne_room_rank = frozenset([
    RoomRank.four_player_throne_east, RoomRank.four_player_throne_south
])
all_three_player_throne_room_rank = frozenset([
    RoomRank.three_player_throne_east, RoomRank.three_player_throne_south
])



import re
from datetime import datetime, timezone, timedelta
from monthdelta import monthdelta
from typing import Optional, Tuple

from nonebot_plugin_majsoul.errors import BadRequestError
from nonebot_plugin_majsoul.utils.decode_integer import decode_integer


def try_parse_time_span(raw: str) -> Optional[Tuple[datetime, datetime]]:
    result = re.match(r"^最近(.*)((天)|(周)|(个月)|(年))$", raw)
    if result is not None:
        num = result.group(1)
        try:
            num = decode_integer(num)
        except ValueError:
            return None

        if num <= 0:
            raise BadRequestError("时间不合法")

        end_time = datetime.now(timezone.utc)

        unit = result.group(2)
        if unit == '天':
            start_time = end_time - timedelta(days=num)
        elif unit == '周':
            start_time = end_time - timedelta(days=num * 7)
        elif unit == '个月':
            start_time = end_time - monthdelta(num)
        elif unit == '年':
            start_time = end_time - monthdelta(num * 12)
        else:
            return None

        return start_time, end_time

    return None



from typing import Optional

from nonebot import get_driver, logger
from tensoul import MajsoulPaipuDownloader
from websockets.exceptions import ConnectionClosedError

from ..config import conf
from ..network.auto_retry import auto_retry

_downloader: Optional[MajsoulPaipuDownloader] = None


def get_downloader():
    if _downloader is None:
        raise RuntimeError("downloader is not ready")
    return _downloader


@get_driver().on_startup
async def restart_downloader():
    global _downloader

    if _downloader is not None:
        await _downloader.close()
        _downloader = None

    if conf.majsoul_username:
        _downloader = MajsoulPaipuDownloader()
        await _downloader.start()
        await _downloader.login(conf.majsoul_username, conf.majsoul_password)
        logger.opt(colors=True).success(f"Logged in as <y>{conf.majsoul_username}</y>")


@auto_retry(ConnectionClosedError, restart_downloader)
async def download_paipu(uuid: str):
    return await get_downloader().download(uuid)


@get_driver().on_shutdown
async def _destroy_downloader():
    global _downloader

    if _downloader is not None:
        await _downloader.close()
        _downloader = None


__all__ = ("get_downloader", "restart_downloader", "download_paipu")



from typing import List

from pydantic import BaseModel, Extra, root_validator

from .room_rank import RoomRank


# noinspection NonAsciiCharacters
class PlayerExtendedStats(BaseModel):
    id: int
    count: int
    played_modes: List[RoomRank]
    和牌率: float = 0.0
    自摸率: float = 0.0
    默听率: float = 0.0
    放铳率: float = 0.0
    副露率: float = 0.0
    立直率: float = 0.0
    平均打点: int = 0
    # 最大连庄: int
    和了巡数: float = 0.0
    平均铳点: int = 0
    流局率: float = 0.0
    流听率: float = 0.0
    # 一发率: float
    # 里宝率: float
    # 被炸率: float
    # 平均被炸点数: int
    # 放铳时立直率: float
    # 放铳时副露率: float
    # 立直后放铳率: float
    # 立直后非瞬间放铳率: float
    # 副露后放铳率: float
    # 立直后和牌率: float
    # 副露后和牌率: float
    # 立直后流局率: float
    # 副露后流局率: float
    # 放铳至立直: int
    # 放铳至副露: int
    # 放铳至默听: int
    # 立直和了: int
    # 副露和了: int
    # 默听和了: int
    立直巡目: float = 0.0
    # 立直收支: int
    # 立直收入: int
    # 立直支出: int
    先制率: float = 0.0
    追立率: float = 0.0
    # 被追率: float
    # 振听立直率: float
    # 立直好型: float
    # 立直多面: float
    # 立直好型2: float
    # 最大累计番数: int
    # W立直: int
    打点效率: int = 0
    # 铳点损失: int
    净打点效率: int = 0

    # 平均起手向听: float
    # 平均起手向听亲: float
    # 平均起手向听子: float

    @root_validator(pre=True, allow_reuse=True)
    def filter_not_none(cls, values):
        for k in list(values.keys()):
            if values[k] is None:
                del values[k]
        return values

    class Config:
        extra = Extra.ignore



from typing import TextIO, Optional

import tzlocal

from nonebot_plugin_majsoul.paifuya.data.models.game_record import GameRecord
from nonebot_plugin_majsoul.paifuya.mappers.player_rank import map_player_rank
from nonebot_plugin_majsoul.paifuya.mappers.room_rank import map_room_rank
from nonebot_plugin_majsoul.utils.rank import ranked

_local_timezone = tzlocal.get_localzone()


def _encode_account_id(account_id: int) -> int:
    return 1358437 + ((7 * account_id + 1117113) ^ 86216345)


def _get_record_link(record: GameRecord, player_id: int) -> str:
    # https://github.com/SAPikachu/amae-koromo/blob/7b99f10f0fce251641013c5571c4fbe5edaa20f1/src/data/types/record.ts#L56
    #
    # encodeAccountId: (t: number) => 1358437 + ((7 * t + 1117113) ^ 86216345),
    #
    # // ...
    #
    # getRecordLink(rec: GameRecord | string, player?: PlayerRecord | number | string) {
    #     const playerId = typeof player === "object" ? player.accountId : player;
    #     const trailer = playerId
    #       ? `_a${GameRecord.encodeAccountId(typeof playerId === "number" ? playerId : parseInt(playerId))}`
    #       : "";
    #     const uuid = typeof rec === "string" ? rec : rec.uuid;
    #     return `${i18n.t("https://game.maj-soul.net/1/")}?paipu=${uuid}${trailer}`;
    #   }

    trailer = f"_a{_encode_account_id(player_id)}"
    return f"https://game.maj-soul.net/1/?paipu={record.uuid}{trailer}"


def map_game_record(sio: TextIO, record: GameRecord, player_id: int, num: Optional[int] = None):
    if num is not None:
        sio.write(f"[{num}]  ")

    for rank, player in ranked(record.players, key=lambda x: x.pt, reverse=True):
        if player.id == player_id:
            sio.write(f"[#{rank}]")
        else:
            sio.write(f"#{rank}")
        sio.write(f'  [{map_player_rank(player.rank)}]{player.nickname}    {player.score}  ')

        sio.write('(')
        if player.pt > 0:
            sio.write('+')
        elif player.pt == 0:
            sio.write('±')
        sio.write(str(player.pt))
        sio.write(')\n')

    sio.write("\n")
    sio.write(f"房间：{map_room_rank({record.room_rank})}\n")
    sio.write(f"时间：{record.start_time.astimezone(_local_timezone).strftime('%Y/%m/%d %H:%M')}\n")
    sio.write(_get_record_link(record, player_id))
    sio.write("\n")



from typing import TextIO, AbstractSet, Union

from .player_rank import map_player_rank
from .room_rank import map_room_rank
from ..data.models.player_info import PlayerLevel
from ..data.models.player_num import PlayerNum
from ..data.models.player_rank import PlayerRank, PlayerMajorRank
from ..data.models.player_stats import PlayerStats
from ..data.models.room_rank import RoomRank
from ...utils.percentile import percentile_str


def map_player_stats(sio: TextIO, stats: PlayerStats,
                     room_rank: Union[AbstractSet[RoomRank], str],
                     player_num: PlayerNum):
    if not isinstance(room_rank, str):
        room_rank = map_room_rank(room_rank)

    sio.write(f"【{room_rank} 基础数据】\n")

    level = fix_level(stats.level)
    sio.write(f"记录段位：{map_player_rank(level.id)}    ")
    sio.write(f"记录PT：{level.score}\n")

    sio.write(f"总场次：{stats.count}    ")
    sio.write(f"平均顺位：{round(stats.avg_rank, 2)}\n")

    sio.write(f"一位率：{percentile_str(stats.rank_rates[0])}    ")
    sio.write(f"二位率：{percentile_str(stats.rank_rates[1])}\n")

    sio.write(f"三位率：{percentile_str(stats.rank_rates[2])}")
    if player_num == PlayerNum.four:
        sio.write("    ")
        sio.write(f"四位率：{percentile_str(stats.rank_rates[3])}")
    sio.write("\n")


def fix_level(level: PlayerLevel) -> PlayerLevel:
    pt = level.score + level.delta
    if pt < 0:
        # 已掉段
        if level.id.minor_rank == 1:
            rank = PlayerRank(level.id.player_num,
                              PlayerMajorRank(level.id.major_rank.value - 1),
                              3)
        else:
            rank = PlayerRank(level.id.player_num,
                              level.id.major_rank,
                              level.id.minor_rank - 1)
        pt = rank.max_pt // 2
    elif pt > level.id.max_pt:
        # 已升段
        if level.id.major_rank != PlayerMajorRank.celestial and level.id.minor_rank == 3:
            rank = PlayerRank(level.id.player_num,
                              PlayerMajorRank(level.id.major_rank.value + 1),
                              1)
        else:
            rank = PlayerRank(level.id.player_num,
                              level.id.major_rank,
                              level.id.minor_rank + 1)
        pt = rank.max_pt // 2
    else:
        rank = level.id

    return PlayerLevel(id=rank, score=pt, delta=0)



from enum import Enum
from functools import total_ordering
from typing import NamedTuple

from .player_num import PlayerNum


class PlayerMajorRank(int, Enum):
    novice = 1
    adept = 2
    expert = 3
    master = 4
    saint = 5
    celestial_old = 6
    celestial = 7


_RANK_MAX_PT = {
    203: 1000,
    301: 1200,
    302: 1400,
    303: 2000,
    401: 2800,
    402: 3200,
    403: 3600,
    501: 4000,
    502: 6000,
    503: 9000,
}

_CELESTIAL_MAX_PT = 10000


@total_ordering
class PlayerRank(NamedTuple):
    player_num: PlayerNum
    major_rank: PlayerMajorRank
    minor_rank: int

    @property
    def code(self) -> int:
        return self.player_num.value * 10000 + self.major_rank * 100 + self.minor_rank

    @classmethod
    def from_code(cls, code: int):
        game_players = code // 10000
        major_rank = code % 1000 // 100
        minor_rank = code % 100
        return PlayerRank(PlayerNum(game_players), PlayerMajorRank(major_rank), minor_rank)

    @property
    def max_pt(self) -> int:
        if self.major_rank != PlayerMajorRank.celestial and self.major_rank != PlayerMajorRank.celestial_old:
            return _RANK_MAX_PT[self.major_rank * 100 + self.minor_rank]
        else:
            return _CELESTIAL_MAX_PT

    def __lt__(self, other):
        if isinstance(other, PlayerRank):
            return self.code < other.code
        else:
            raise TypeError(type(other))


__all__ = ("PlayerMajorRank", "PlayerRank")



_DIGIT_MAPPING = {
    1: "一",
    2: "二",
    3: "三",
    4: "四",
    5: "五",
    6: "六",
    7: "七",
    8: "八",
    9: "九",
    10: "十"
}


def map_digit(digit: int) -> str:
    return _DIGIT_MAPPING.get(digit, str(digit))



from ..data.models.player_rank import PlayerRank, PlayerMajorRank

_MAJOR_RANK_MAPPING = {
    PlayerMajorRank.novice: "心",
    PlayerMajorRank.adept: "士",
    PlayerMajorRank.expert: "杰",
    PlayerMajorRank.master: "豪",
    PlayerMajorRank.saint: "圣",
    PlayerMajorRank.celestial_old: "魂",
    PlayerMajorRank.celestial: "魂",
}


def map_player_rank(rank: PlayerRank) -> str:
    return _MAJOR_RANK_MAPPING[rank.major_rank] + str(rank.minor_rank)



from typing import TypeVar, Iterable, Tuple, Optional, Callable, Any

T = TypeVar("T")


def ranked(__iterable: Iterable[T],
           *, key: Optional[Callable[[T], Any]] = None,
           reverse: bool = False) -> Iterable[Tuple[int, T]]:
    if key is None:
        key = lambda x: x

    rank = 0
    prev = None
    for x in sorted(__iterable, key=key, reverse=reverse):
        if prev is None or key(x) != prev:
            rank += 1
        yield rank, x



from functools import wraps
from typing import TypeVar, Callable, Optional, Awaitable, Any, Union, Type, Sequence

from nonebot import logger
from typing_extensions import ParamSpec

T = TypeVar('T')
P = ParamSpec('P')


def auto_retry(error: Union[Type[Exception], Sequence[Type[Exception]]],
               before_retry: Optional[Callable[[], Awaitable[Any]]] = None):
    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        @wraps(func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            err = None

            for i in range(10):
                try:
                    return await func(*args, **kwargs)
                except error as e:
                    if before_retry:
                        await before_retry()
                    logger.error(f"Retrying... {i + 1}/10")
                    logger.exception(e)
                    err = e
                except Exception as e:
                    raise e

            raise err

        return wrapper

    return decorator



from datetime import datetime, timedelta
from functools import partial
from typing import List, AbstractSet, AsyncGenerator

from httpx import AsyncClient, URL, HTTPError
from nonebot import get_driver, logger

from nonebot_plugin_majsoul.network.auto_retry import auto_retry
from nonebot_plugin_majsoul.network.host_prober import HostProber
from .models.game_record import GameRecord
from .models.player_extended_stats import PlayerExtendedStats
from .models.player_info import PlayerInfo
from .models.player_num import PlayerNum
from .models.player_stats import PlayerStats
from .models.room_rank import RoomRank

prober = HostProber([
    "1.data.amae-koromo.com",
    "2.data.amae-koromo.com",
    "3.data.amae-koromo.com",
    "4.data.amae-koromo.com",
    "5-data.amae-koromo.com",
])


class PaifuyaApi:
    def __init__(self, baseurl):
        self._baseurl = baseurl

        async def req_hook(request):
            logger.trace(f"Request: {request.method} {request.url} - Waiting for response")

        async def resp_hook(response):
            request = response.request
            logger.trace(f"Response: {request.method} {request.url} - Status {response.status_code}")
            response.raise_for_status()

        self.client: AsyncClient = AsyncClient(
            follow_redirects=True,
            event_hooks={'request': [req_hook], 'response': [resp_hook]}
        )

        get_driver().on_shutdown(self.close)

    async def close(self):
        await self.client.aclose()

    @auto_retry(HTTPError, before_retry=partial(prober.select_host, exclude_current=True))
    async def search_player(
            self, nickname: str,
            *, limit: int = 10
    ) -> List[PlayerInfo]:
        resp = await self.client.get(
            URL(f"https://{prober.host}/{self._baseurl}/search_player/{nickname}"),
            params={"limit": limit}
        )
        return [PlayerInfo.parse_obj(x) for x in resp.json()]

    @auto_retry(HTTPError, before_retry=partial(prober.select_host, exclude_current=True))
    async def player_stats(
            self, player_id: int, start_time: datetime, end_time: datetime, room_rank: AbstractSet[RoomRank]
    ) -> PlayerStats:
        start_timestamp = int(start_time.timestamp() * 1000)
        end_timestamp = int(end_time.timestamp() * 1000)
        mode = ".".join(map(lambda x: str(x.value), room_rank))
        resp = await self.client.get(
            URL(f"https://{prober.host}/{self._baseurl}/player_stats/{player_id}/{start_timestamp}/{end_timestamp}"),
            params={"mode": mode}
        )
        return PlayerStats.parse_obj(resp.json())

    @auto_retry(HTTPError, before_retry=partial(prober.select_host, exclude_current=True))
    async def player_extended_stats(
            self, player_id: int, start_time: datetime, end_time: datetime, room_rank: AbstractSet[RoomRank]
    ) -> PlayerExtendedStats:
        start_timestamp = int(start_time.timestamp() * 1000)
        end_timestamp = int(end_time.timestamp() * 1000)
        mode = ".".join(map(lambda x: str(x.value), room_rank))
        resp = await self.client.get(
            URL(f"https://{prober.host}/{self._baseurl}/player_extended_stats/{player_id}/{start_timestamp}/{end_timestamp}"),
            params={"mode": mode}
        )
        return PlayerExtendedStats.parse_obj(resp.json())

    @auto_retry(HTTPError, before_retry=partial(prober.select_host, exclude_current=True))
    async def player_records(
            self, player_id: int, start_time: datetime, end_time: datetime, room_rank: AbstractSet[RoomRank],
            *, limit: int, descending: bool = True
    ) -> List[GameRecord]:
        start_timestamp = int(start_time.timestamp() * 1000)
        end_timestamp = int(end_time.timestamp() * 1000)
        mode = ".".join(map(lambda x: str(x.value), room_rank))
        resp = await self.client.get(
            URL(f"https://{prober.host}/{self._baseurl}/player_records/{player_id}/{end_timestamp}/{start_timestamp}"),
            params={"mode": mode, "limit": str(limit), "descending": str(descending).lower()}
        )
        return [GameRecord.parse_obj(x) for x in resp.json()]

    async def player_records_stream(
            self, player_id: int, start_time: datetime, end_time: datetime, room_rank: AbstractSet[RoomRank],
            *, batch: int = 200, descending: bool = True
    ) -> AsyncGenerator[GameRecord, None]:
        while start_time <= end_time:
            records = await self.player_records(
                player_id, start_time, end_time, room_rank, limit=batch, descending=descending
            )
            for r in records:
                yield r

            if len(records) < batch:
                break

            end_time = records[-1].start_time - timedelta(seconds=1)


four_player_api = PaifuyaApi(f"api/v2/pl4")
three_player_api = PaifuyaApi(f"api/v2/pl3")

paifuya_api = {
    PlayerNum.four: four_player_api,
    PlayerNum.three: three_player_api,
}

__all__ = ("PaifuyaApi", "four_player_api", "three_player_api", "paifuya_api")



from datetime import datetime, timezone
from typing import List

from pydantic import BaseModel, Field, validator

from .player_rank import PlayerRank
from .room_rank import RoomRank


class GamePlayer(BaseModel):
    id: int = Field(alias="accountId")
    nickname: str
    rank: PlayerRank = Field(alias="level")
    score: int
    pt: int = Field(alias="gradingScore")

    @validator("rank", pre=True, allow_reuse=True)
    def parse_rank(cls, v):
        if isinstance(v, PlayerRank):
            return v
        elif isinstance(v, int):
            return PlayerRank.from_code(v)
        else:
            raise TypeError(f"value must be int or PlayerRank, got {type(v)}")


class GameRecord(BaseModel):
    room_rank: RoomRank = Field(alias="modeId")
    uuid: str
    start_time: datetime = Field(alias="startTime")
    end_time: datetime = Field(alias="endTime")
    players: List[GamePlayer]

    @validator("start_time", "end_time", pre=True, allow_reuse=True)
    def parse_timestamp(cls, v):
        if isinstance(v, int) or isinstance(v, float):
            return datetime.fromtimestamp(v, timezone.utc)
        else:
            return v



from pydantic import BaseModel, validator, Field

from .player_rank import PlayerRank


class PlayerLevel(BaseModel):
    id: PlayerRank
    score: int
    delta: int

    @validator("id", pre=True, allow_reuse=True)
    def parse_rank(cls, v):
        if isinstance(v, PlayerRank):
            return v
        elif isinstance(v, int):
            return PlayerRank.from_code(v)
        else:
            raise TypeError(f"value must be int or PlayerRank, got {type(v)}")


class PlayerInfo(BaseModel):
    id: int
    nickname: str
    level: PlayerLevel
    latest_timestamp: int



from typing import TextIO, Union, AbstractSet

from nonebot_plugin_majsoul.utils.percentile import percentile_str
from .room_rank import map_room_rank
from ..data.models.player_extended_stats import PlayerExtendedStats
from ..data.models.room_rank import RoomRank


def map_player_extended_stats(sio: TextIO, stats: PlayerExtendedStats,
                              room_rank: Union[AbstractSet[RoomRank], str]):
    if not isinstance(room_rank, str):
        room_rank = map_room_rank(room_rank)

    sio.write(f"【{room_rank} 详细数据】\n")

    sio.write(f"和牌率：{percentile_str(stats.和牌率)}    ")
    sio.write(f"放铳率：{percentile_str(stats.放铳率)}    ")
    sio.write(f"自摸率：{percentile_str(stats.自摸率)}    ")
    sio.write(f"默胡率：{percentile_str(stats.默听率)}\n")

    sio.write(f"副露率：{percentile_str(stats.副露率)}    ")
    sio.write(f"立直率：{percentile_str(stats.立直率)}    ")
    sio.write(f"流局率：{percentile_str(stats.流局率)}    ")
    sio.write(f"流听率：{percentile_str(stats.流听率)}\n")

    sio.write(f"先制率：{percentile_str(stats.先制率)}    ")
    sio.write(f"追立率：{percentile_str(stats.追立率)}    ")
    sio.write(f"立直巡目：{round(stats.立直巡目, 2)}    ")
    sio.write(f"和了巡目：{round(stats.和了巡数, 2)}\n")

    sio.write(f"平均打点：{stats.平均打点}    ")
    sio.write(f"平均铳点：{stats.平均铳点}    ")
    sio.write(f"打点效率：{stats.打点效率}    ")
    sio.write(f"净打点效率：{stats.净打点效率}\n")



from nonebot import logger

from .downloader import download_paipu
from ..config import conf

if not conf.majsoul_username:
    logger.warning("majsoul_paipu is disabled because majsoul_username is not configured")
else:
    import json
    import re
    from asyncio import wait_for

    from nonebot import on_command, logger, require
    from nonebot.adapters.onebot.v11 import Bot, GroupMessageEvent, MessageEvent
    from nonebot.params import CommandArg
    from tensoul.downloader import MajsoulDownloadError

    from ..errors import BadRequestError
    from ..interceptors.handle_error import handle_error
    from ..utils.nonebot import default_cmd_start

    uuid_reg = re.compile(r"\d{6}-[\da-fA-F]{8}-[\da-fA-F]{4}-[\da-fA-F]{4}-[\da-fA-F]{4}-[\da-fA-F]{12}")

    require("nonebot_plugin_gocqhttp_cross_machine_upload_file")

    from nonebot_plugin_gocqhttp_cross_machine_upload_file import upload_group_file, upload_private_file

    query_majsoul_paipu_matcher = on_command("下载雀魂牌谱")


    @query_majsoul_paipu_matcher.handle()
    @handle_error(query_majsoul_paipu_matcher)
    async def majsoul_paipu(bot: Bot, event: MessageEvent, args=CommandArg()):
        plain_args = args.extract_plain_text()
        mat = uuid_reg.search(plain_args)
        if not mat:
            raise BadRequestError(f"使用方式：{default_cmd_start}下载雀魂牌谱 <牌谱网址>")

        uuid = mat.group(0)

        logger.opt(colors=True).info(f"Downloading paipu <y>{uuid}</y>")
        try:
            coro = download_paipu(uuid)
            if conf.majsoul_query_timeout:
                record = await wait_for(coro, timeout=conf.majsoul_query_timeout)
            else:
                record = await coro
        except MajsoulDownloadError as e:
            logger.opt(colors=True).warning(f"Failed to download paipu <y>{uuid}</y>, code: {e.code}")
            if e.code == 1203:
                raise BadRequestError("牌谱不存在") from e
            else:
                raise e

        time_str = record['title'][1]
        year = int(time_str[0:4])
        month = int(time_str[5:7])
        day = int(time_str[8:10])
        hour = int(time_str[11:13])
        minute = int(time_str[14:16])

        filename = f"{record['title'][0]}_{year}_{month}_{day}_{hour}_{minute}（{'、'.join(record['name'])}）.json"

        data = json.dumps(record, ensure_ascii=False).encode("utf-8")

        if isinstance(event, GroupMessageEvent):
            await upload_group_file(bot, event.group_id, filename, data)
        else:
            await upload_private_file(bot, event.user_id, filename, data)



from nonebot import get_driver
from pydantic import BaseSettings


class Config(BaseSettings):
    majsoul_query_timeout: float = 15.0

    majsoul_username: str = ""
    majsoul_password: str = ""

    majsoul_font: str = ""
    majsoul_font_path: str = ""

    class Config:
        extra = "ignore"


conf = Config(**get_driver().config.dict())



from asyncio import create_task, sleep
from typing import TypeVar, Sequence

from icmplib import async_ping, NameLookupError, SocketAddressError, ICMPSocketError
from nonebot import get_driver, logger
from typing_extensions import ParamSpec

from ..utils.percentile import percentile_str

T = TypeVar('T')
P = ParamSpec('P')


class HostProber:
    def __init__(self, mirrors: Sequence[str]):
        self._mirrors = mirrors
        self._host = mirrors[0]
        self._select_host_daemon_task = None

        get_driver().on_startup(self.start)
        get_driver().on_shutdown(self.close)

    async def start(self):
        self._select_host_daemon_task = create_task(self._select_host_daemon())

    async def close(self):
        self._select_host_daemon_task.cancel()

    @property
    def host(self) -> str:
        return self._host

    async def select_host(self, exclude_current: bool = False) -> bool:
        logger.debug("paifuya host selecting...")

        ping_tasks = [create_task(async_ping(h))
                      if not exclude_current or h == self._host
                      else None
                      for h in self._mirrors]

        selected = None
        selected_result = None

        for i, ping_task in enumerate(ping_tasks):
            if ping_task is None:
                continue

            try:
                ping_result = await ping_task
            except (NameLookupError, SocketAddressError, ICMPSocketError) as e:
                logger.trace(f"failed to ping to {self._mirrors[i]} ({str(type(e))})")
                continue

            logger.trace(f"ping to {self._mirrors[i]}")
            logger.trace(ping_result)

            if ping_result.packet_loss == 1:
                logger.trace(f"{self._mirrors[i]} was ignored due to all packages loss")
                continue

            if (selected is None or
                    (ping_result.packet_loss, ping_result.avg_rtt) <
                    (ping_result.packet_loss, selected_result.avg_rtt)):
                selected = self._mirrors[i]
                selected_result = ping_result

        if selected is not None:
            self._host = selected
            logger.info(f"switched paifuya host to {selected}  "
                        f"(avg rtt: {selected_result.avg_rtt}ms, "
                        f"packet loss: {percentile_str(selected_result.packet_loss)})")
            return True
        else:
            return False

    async def _select_host_daemon(self):
        try:
            while True:
                if await self.select_host():
                    await sleep(10 * 60)
                else:
                    logger.error(f"all ping to paifuya host has failed. will retry after 60s...")
                    await sleep(60)
        except Exception as e:
            logger.exception(e)



from asyncio import wait_for
from datetime import datetime, timezone
from io import StringIO, BytesIO
from typing import Optional, AbstractSet, Sequence, Text, List

import matplotlib.pyplot as plt
from httpx import HTTPStatusError
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from nonebot import on_command
from nonebot.internal.adapter import Event
from nonebot.internal.matcher import current_bot
from nonebot_plugin_saa import MessageFactory, AggregatedMessageFactory, Image

from nonebot_plugin_majsoul.config import conf
from nonebot_plugin_majsoul.errors import BadRequestError
from nonebot_plugin_majsoul.interceptors.handle_error import handle_error
from .data.api import paifuya_api as api
from .data.models.game_record import GameRecord
from .data.models.player_num import PlayerNum
from .data.models.room_rank import RoomRank, all_four_player_room_rank, all_three_player_room_rank
from .mappers.game_record import map_game_record
from .mappers.room_rank import map_room_rank
from .parsers.room_rank import try_parse_room_rank
from ..utils.my_executor import run_in_my_executor
from ..utils.rank import ranked


def make_handler(player_num: PlayerNum):
    async def majsoul_records(event: Event):
        args = event.get_message().extract_plain_text().split()
        cmd, args = args[0], args[1:]

        if len(args) == 0:
            raise BadRequestError(f"指令格式：{cmd} <雀魂账号> [<房间类型>]")

        nickname = args[0]
        if len(nickname) > 15:
            raise BadRequestError("昵称长度超过雀魂最大限制")

        kwargs = {}

        for arg in args[1:]:
            if "room_rank" not in kwargs:
                room_rank = try_parse_room_rank(arg)
                if room_rank is not None:
                    if player_num == PlayerNum.four:
                        kwargs["room_rank"] = room_rank[0]
                    elif player_num == PlayerNum.three:
                        kwargs["room_rank"] = room_rank[1]
                    continue

        coro = handle_majsoul_records(nickname, player_num, **kwargs)
        if conf.majsoul_query_timeout:
            await wait_for(coro, timeout=conf.majsoul_query_timeout)
        else:
            await coro

    return majsoul_records


four_player_majsoul_records_matcher = on_command("雀魂最近对局", aliases={'雀魂对局', '雀魂牌谱'})
four_player_majsoul_records = make_handler(PlayerNum.four)
four_player_majsoul_records = handle_error(four_player_majsoul_records_matcher)(
    four_player_majsoul_records)
four_player_majsoul_records_matcher.append_handler(four_player_majsoul_records)

three_player_majsoul_records_matcher = on_command("雀魂三麻最近对局", aliases={'雀魂三麻对局', '雀魂三麻牌谱'})
three_player_majsoul_records = make_handler(PlayerNum.three)
three_player_majsoul_records = handle_error(three_player_majsoul_records_matcher)(
    three_player_majsoul_records)
three_player_majsoul_records_matcher.append_handler(three_player_majsoul_records)


def draw_records_plot(bio: BytesIO, records: Sequence[GameRecord], player_id: int):
    fig: Figure = plt.figure(facecolor='w', figsize=(12, 2))
    ax: Axes = fig.add_subplot(1, 1, 1)

    rank = []
    for r in records:
        for i, p in ranked(r.players, key=lambda x: x.pt, reverse=True):
            if p.id != player_id:
                continue

            rank.append(i)
    rank.reverse()
    ax.plot(range(1, len(records) + 1), rank, marker='o')

    # 设置X轴与Y轴范围
    ax.set_yticks([1, 2, 3, 4], labels=[1, 2, 3, 4])
    ax.set_xlim(0.5, len(records) + 0.5)
    ax.set_ylim(0.75, 4.25)

    # 反转Y轴
    ax.invert_yaxis()

    # 移除边框
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['left'].set_visible(False)

    # 显示网格
    ax.grid(axis='y')

    # 隐藏X轴
    ax.get_xaxis().set_visible(False)

    fig.savefig(bio, format='png')


async def handle_majsoul_records(nickname: str, player_num: PlayerNum, *,
                                 room_rank: Optional[AbstractSet[RoomRank]] = None):
    if room_rank is None:
        if player_num == PlayerNum.four:
            room_rank = all_four_player_room_rank
        elif player_num == PlayerNum.three:
            room_rank = all_three_player_room_rank

    start_time = datetime.fromisoformat("2010-01-01T00:00:00")
    end_time = datetime.now(timezone.utc)

    msgs: List[MessageFactory] = []

    players = await api[player_num].search_player(nickname)
    if len(players) == 0:
        msgs.append(MessageFactory(Text("没有查询到该角色在金之间以上的对局数据呢~")))
    else:
        try:
            records = await api[player_num].player_records(
                players[0].id,
                start_time,
                end_time,
                room_rank,
                limit=10,
                descending=True)
        except HTTPStatusError as e:
            if e.response.status_code == 404:
                records = None
            else:
                raise e

        msg = ""
        if len(players) > 1:
            msg += "查询到多条角色昵称呢~，若输出不是您想查找的昵称，请补全查询昵称。\n"
        msg += f"昵称：{players[0].nickname}"

        room_rank_text = map_room_rank(room_rank)
        if not records:
            msg += f"\n\n没有查询到{room_rank_text}的对局数据呢~"

        msgs.append(MessageFactory(Text(msg.strip())))

        if records:
            with BytesIO() as bio:
                await run_in_my_executor(draw_records_plot, bio, records, players[0].id)
                msgs.append(MessageFactory(Image(bio.getvalue())))

            for i, r in enumerate(records):
                with StringIO() as sio:
                    map_game_record(sio, r, players[0].id)
                    msgs.append(MessageFactory(Text(sio.getvalue().strip())))

            with StringIO() as url:
                if player_num == PlayerNum.four:
                    url.write(f"https://amae-koromo.sapk.ch/player/{players[0].id}/")
                else:
                    url.write(f"https://ikeda.sapk.ch/player/{players[0].id}/")
                url.write(".".join(map(lambda x: str(x.value), room_rank)))

                msgs.append(MessageFactory(Text(f"更多信息：{url.getvalue()}")))

    if len(msgs) == 1:
        await msgs[0].send(reply=True)
    else:
        adapter_name = current_bot.get().adapter.get_name()

        if adapter_name in AggregatedMessageFactory.sender:
            await AggregatedMessageFactory(msgs).send()
        else:
            for msg in msgs:
                await msg.send(reply=True)



import re
from typing import Optional

from nonebot_plugin_majsoul.errors import BadRequestError
from nonebot_plugin_majsoul.utils.decode_integer import decode_integer


def try_parse_limit_of_games(raw: str) -> Optional[int]:
    result = re.match(r"^最近(.*)场$$", raw)
    if result is not None:
        num = result.group(1)
        try:
            num = decode_integer(num)
        except ValueError:
            return None

        if num <= 0:
            raise BadRequestError("场数不合法")

        return num

    return None



import sys
from asyncio import wait_for
from datetime import datetime, timezone
from io import BytesIO
from typing import Sequence, Optional, Text

import matplotlib.pyplot as plt
from httpx import HTTPStatusError
from matplotlib import font_manager
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from nonebot import on_command
from nonebot.internal.adapter import Event
from nonebot_plugin_saa import MessageFactory, Image

from nonebot_plugin_majsoul.config import conf
from nonebot_plugin_majsoul.errors import BadRequestError
from nonebot_plugin_majsoul.interceptors.handle_error import handle_error
from .data.api import paifuya_api as api
from .data.models.game_record import GameRecord
from .data.models.player_info import PlayerInfo, PlayerLevel
from .data.models.player_num import PlayerNum
from .data.models.player_rank import PlayerMajorRank
from .data.models.room_rank import all_four_player_room_rank, all_three_player_room_rank, RoomRank
from .mappers.player_num import map_player_num
from .mappers.player_rank import map_player_rank
from .parsers.limit_of_games import try_parse_limit_of_games
from .parsers.time_span import try_parse_time_span
from ..utils.my_executor import run_in_my_executor

if conf.majsoul_font:
    plt.rcParams['font.sans-serif'] = conf.majsoul_font
elif sys.platform == 'win32':
    plt.rcParams['font.sans-serif'] = "Microsoft YaHei"

if conf.majsoul_font_path:
    font_manager.fontManager.addfont(conf.majsoul_font_path)


def make_handler(player_num: PlayerNum):
    async def majsoul_pt_plot(event: Event):
        args = event.get_message().extract_plain_text().split()
        cmd, args = args[0], args[1:]

        if len(args) == 0:
            raise BadRequestError(f"指令格式：{cmd} <雀魂账号> [最近<数量>场] [最近<数量>{{天|周|个月|年}}]")

        nickname = args[0]
        if len(nickname) > 15:
            raise BadRequestError("昵称长度超过雀魂最大限制")

        kwargs = {}

        for arg in args[1:]:
            if "time_span" not in kwargs:
                time_span = try_parse_time_span(arg)
                if time_span is not None:
                    kwargs["start_time"], kwargs["end_time"] = time_span
                    continue

            if "limit" not in kwargs:
                limit = try_parse_limit_of_games(arg)
                if limit is not None:
                    kwargs["limit"] = limit
                    continue

        coro = handle_majsoul_pt_plot(nickname, player_num, **kwargs)
        if conf.majsoul_query_timeout:
            await wait_for(coro, timeout=conf.majsoul_query_timeout)
        else:
            await coro

    return majsoul_pt_plot


four_player_majsoul_pt_plot_matcher = on_command("雀魂PT推移图", aliases={"雀魂PT图"})
four_player_majsoul_pt_plot_records = make_handler(PlayerNum.four)
four_player_majsoul_pt_plot_records = handle_error(four_player_majsoul_pt_plot_matcher)(
    four_player_majsoul_pt_plot_records)
four_player_majsoul_pt_plot_matcher.append_handler(four_player_majsoul_pt_plot_records)

three_player_majsoul_pt_plot_matcher = on_command("雀魂三麻PT推移图", aliases={"雀魂三麻PT图"})
three_player_majsoul_pt_plot_records = make_handler(PlayerNum.three)
three_player_majsoul_pt_plot_records = handle_error(three_player_majsoul_pt_plot_matcher)(
    three_player_majsoul_pt_plot_records)
three_player_majsoul_pt_plot_matcher.append_handler(three_player_majsoul_pt_plot_records)

_color = {RoomRank.four_player_throne_south: 'r',
          RoomRank.four_player_throne_east: 'r',
          RoomRank.four_player_jade_south: 'g',
          RoomRank.four_player_jade_east: 'g',
          RoomRank.four_player_golden_south: 'y',
          RoomRank.four_player_golden_east: 'y',
          RoomRank.three_player_throne_south: 'r',
          RoomRank.three_player_throne_east: 'r',
          RoomRank.three_player_jade_south: 'g',
          RoomRank.three_player_jade_east: 'g',
          RoomRank.three_player_golden_south: 'y',
          RoomRank.three_player_golden_east: 'y'}


def draw(bio: BytesIO,
         player_num: PlayerNum,
         player: PlayerInfo,
         initial_level: PlayerLevel,
         records: Sequence[GameRecord]):
    fig: Figure = plt.figure(facecolor='w', figsize=(16, 10))
    ax: Axes = fig.add_subplot(1, 1, 1)

    pre_rank = max_rank = initial_level.id
    pre_pt = pt = initial_level.score + initial_level.delta
    base = pre_rank.max_pt // 2

    ax.text(3, 100, '\n'.join(map_player_rank(pre_rank)), fontsize=15)

    is_celestial = False

    for i, r in enumerate(records):
        for p in r.players:
            if p.id != player.id:
                continue

            rank = p.rank
            if max_rank is None or rank > max_rank:
                max_rank = rank

            if pre_rank != rank:
                ax.text(i + 3, 100, '\n'.join(map_player_rank(rank)), fontsize=15)
                ax.vlines(i, 0, max(rank.max_pt, pre_rank.max_pt if pre_rank else 0), color='k')

                base = rank.max_pt // 2
                pt = pre_pt = base

            if rank.major_rank == PlayerMajorRank.celestial:
                # 获取到的PT是魂珠*100（每20魂珠升一段），这里再缩放5倍
                pt += p.pt * 5
                is_celestial = True
            else:
                pt += p.pt

            ax.plot([i, i + 1], [pre_pt, pt], color='k', lw=1.5)
            ax.fill_between([i, i + 1], [pre_pt, pt], color=_color[r.room_rank], alpha=0.05)
            ax.plot([i, i + 1], [base, base], color='k', lw=1.5)
            ax.plot([i, i + 1], [base * 2, base * 2], color='k', lw=1.5)

            pre_rank, pre_pt = rank, pt

    ax.set_title(f'雀魂段位战PT推移图[{map_player_num(player_num)}]  '
                 f'@{player.nickname}'
                 f'（{records[0].start_time.strftime("%Y/%m/%d")}~{records[-1].start_time.strftime("%Y/%m/%d")}）',
                 fontsize=12, pad=5)
    ax.set_xlabel('对局数', fontsize=20)

    ax.set_ylabel('PT', fontsize=20)

    if not is_celestial:
        ax.set_yticks([i * 1000 for i in range(11)], labels=[i * 1000 for i in range(11)])
    else:
        ax.set_yticks([i * 1000 for i in range(11)], labels=[f'{i * 1000} 魂珠{float(i * 2)}' for i in range(11)])

    ax.set_xlim(0, len(records))
    ax.set_ylim(0, max_rank.max_pt + 100)

    fig.savefig(bio, format='png')


async def handle_majsoul_pt_plot(nickname: str, player_num: PlayerNum, *,
                                 start_time: Optional[datetime] = None,
                                 end_time: Optional[datetime] = None,
                                 limit: Optional[int] = None):
    if player_num == PlayerNum.four:
        room_rank = all_four_player_room_rank
    elif player_num == PlayerNum.three:
        room_rank = all_three_player_room_rank
    else:
        raise ValueError(f"invalid player_num: {player_num}")

    if end_time is None:
        end_time = datetime.now(timezone.utc)

    players = await api[player_num].search_player(nickname)
    if len(players) == 0:
        await MessageFactory(
            Text("没有查询到该角色在金之间以上的对局数据呢~")
        ).send(reply=True)
        return

    player = players[0]

    msg = ""
    if len(players) > 1:
        msg += "查询到多条角色昵称呢~，若输出不是您想查找的昵称，请补全查询昵称。\n"
    msg += f"昵称：{player.nickname}\n"

    try:
        records = []

        api_start_time = datetime.fromisoformat("2010-01-01T00:00:00").astimezone(timezone.utc)
        async for r in api[player_num].player_records_stream(
                player.id, api_start_time, end_time, room_rank, descending=True):
            # 比limit多取一个，用于获取在此之前的段位及PT
            if limit is not None and len(records) > limit:
                break

            records.append(r)

            # 同样多取一个，用于获取在此之前的段位及PT
            if start_time is not None and r.start_time < start_time:
                limit = len(records) - 1
                break

        predecessor_record = None
        if limit:
            if len(records) > limit:
                predecessor_record = records[limit]
            records = records[:limit]

        records.reverse()

        if predecessor_record is not None:
            player_stats_at_start = await api[player_num].player_stats(
                player.id, predecessor_record.start_time, predecessor_record.end_time, room_rank)
            initial_level = player_stats_at_start.level
        else:
            initial_level = None
            for p in records[0].players:
                if p.id == player.id:
                    initial_level = PlayerLevel(id=p.rank, score=p.rank.max_pt // 2, delta=0)
                    break
    except HTTPStatusError as e:
        if e.response.status_code == 404:
            await MessageFactory(
                Text("没有查询到该角色在金之间以上的对局数据呢~")
            ).send(reply=True)
            return
        else:
            raise e

    if not records:
        await MessageFactory(
            Text(msg)
        ).send(reply=True)
    else:
        with BytesIO() as bio:
            await run_in_my_executor(draw, bio, player_num, player, initial_level, records)
            await MessageFactory(
                Image(bio)
            ).send()



from asyncio import create_task, wait_for
from datetime import datetime, timezone
from io import StringIO
from typing import AbstractSet, Optional, Text

from httpx import HTTPStatusError
from nonebot import on_command
from nonebot.internal.adapter import Event
from nonebot_plugin_saa import MessageFactory

from nonebot_plugin_majsoul.config import conf
from nonebot_plugin_majsoul.errors import BadRequestError
from nonebot_plugin_majsoul.interceptors.handle_error import handle_error
from .data.api import paifuya_api as api
from .data.models.player_num import PlayerNum
from .data.models.room_rank import all_four_player_room_rank, all_three_player_room_rank, RoomRank
from .mappers.player_extended_stats import map_player_extended_stats
from .mappers.player_stats import map_player_stats
from .mappers.room_rank import map_room_rank
from .parsers.limit_of_games import try_parse_limit_of_games
from .parsers.room_rank import try_parse_room_rank
from .parsers.time_span import try_parse_time_span


def make_handler(player_num: PlayerNum):
    async def majsoul_info(event: Event):
        args = event.get_message().extract_plain_text().split()
        cmd, args = args[0], args[1:]

        if len(args) == 0:
            raise BadRequestError(f"指令格式：{cmd} <雀魂账号> [<房间类型>] [最近<数量>场] [最近<数量>{{天|周|个月|年}}]")

        nickname = args[0]
        if len(nickname) > 15:
            raise BadRequestError("昵称长度超过雀魂最大限制")

        kwargs = {}

        for arg in args[1:]:
            if "room_rank" not in kwargs:
                room_rank = try_parse_room_rank(arg)
                if room_rank is not None:
                    if player_num == PlayerNum.four:
                        kwargs["room_rank"] = room_rank[0]
                    elif player_num == PlayerNum.three:
                        kwargs["room_rank"] = room_rank[1]
                    continue

            if "time_span" not in kwargs:
                time_span = try_parse_time_span(arg)
                if time_span is not None:
                    kwargs["start_time"], kwargs["end_time"] = time_span
                    continue

            if "limit" not in kwargs:
                limit = try_parse_limit_of_games(arg)
                if limit is not None:
                    kwargs["limit"] = limit
                    continue

        coro = handle_majsoul_info(nickname, player_num, **kwargs)
        if conf.majsoul_query_timeout:
            await wait_for(coro, timeout=conf.majsoul_query_timeout)
        else:
            await coro

    return majsoul_info


four_player_majsoul_info_matcher = on_command('雀魂信息', aliases={'雀魂查询'})
four_player_majsoul_info = make_handler(PlayerNum.four)
four_player_majsoul_info = handle_error(four_player_majsoul_info_matcher)(four_player_majsoul_info)
four_player_majsoul_info_matcher.append_handler(four_player_majsoul_info)

three_player_majsoul_info_matcher = on_command('雀魂三麻信息', aliases={'雀魂三麻查询'})
three_player_majsoul_info = make_handler(PlayerNum.three)
three_player_majsoul_info = handle_error(three_player_majsoul_info_matcher)(three_player_majsoul_info)
three_player_majsoul_info_matcher.append_handler(three_player_majsoul_info)


async def handle_majsoul_info(nickname: str, player_num: PlayerNum, *,
                              room_rank: Optional[AbstractSet[RoomRank]] = None,
                              start_time: Optional[datetime] = None,
                              end_time: Optional[datetime] = None,
                              limit: Optional[int] = None):
    default_start_time = start_time is None
    default_end_time = end_time is None
    default_limit = limit is None

    if room_rank is None:
        if player_num == PlayerNum.four:
            room_rank = all_four_player_room_rank
        elif player_num == PlayerNum.three:
            room_rank = all_three_player_room_rank

    if start_time is None:
        start_time = datetime.fromisoformat("2010-01-01T00:00:00").astimezone(timezone.utc)

    if end_time is None:
        end_time = datetime.now(timezone.utc)

    with StringIO() as sio:
        players = await api[player_num].search_player(nickname)
        if len(players) == 0:
            sio.write("没有查询到该角色在金之间以上的对局数据呢~")
        else:
            if len(players) > 1:
                sio.write("查询到多条角色昵称呢~，若输出不是您想查找的昵称，请补全查询昵称。\n")

            sio.write(f"昵称：{players[0].nickname}\n\n")

            try:
                if limit is not None:
                    records = await api[player_num].player_records(
                        players[0].id,
                        start_time,
                        end_time,
                        room_rank,
                        limit=limit,
                        descending=True)
                    start_time = records[-1].start_time

                player_stats = create_task(api[player_num].player_stats(
                    players[0].id,
                    start_time,
                    end_time,
                    room_rank))
                player_extended_stats = create_task(api[player_num].player_extended_stats(
                    players[0].id,
                    start_time,
                    end_time,
                    room_rank))

                player_stats = await player_stats
                player_extended_stats = await player_extended_stats
            except HTTPStatusError as e:
                if e.response.status_code == 404:
                    player_stats = None
                else:
                    raise e

            room_rank_text = map_room_rank(room_rank)
            if player_stats is None:
                sio.write(f"没有查询到{room_rank_text}的对局数据呢~")
            else:
                map_player_stats(sio, player_stats, room_rank_text, player_num)
                sio.write('\n')
                map_player_extended_stats(sio, player_extended_stats, room_rank_text)

            sio.write("\nPS：本数据不包含金之间以下对局以及2019.11.29之前的对局")

            with StringIO() as url:
                if player_num == PlayerNum.four:
                    url.write(f"https://amae-koromo.sapk.ch/player/{players[0].id}/")
                else:
                    url.write(f"https://ikeda.sapk.ch/player/{players[0].id}/")
                url.write(".".join(map(lambda x: str(x.value), room_rank)))
                if not default_start_time:
                    url.write("/")
                    url.write(start_time.strftime("%Y-%m-%d"))
                if not default_end_time:
                    url.write("/")
                    url.write(end_time.strftime("%Y-%m-%d"))
                if not default_limit:
                    url.write(f"?limit={limit}")

                sio.write("\n")
                sio.write("更多信息：")
                sio.write(url.getvalue())

        msg = sio.getvalue()
        await MessageFactory(Text(msg)).send(reply=True)
